
=begin

# Copyright 2010-2011, Gilles Macabies : macgile@gmail.com

# Permission to use, copy, modify, and distribute this software for
# any purpose and without fee is hereby granted, provided that the above
# copyright notice appear in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name        :   Camera Key Maker 1.0
# Description :   A tool to create scenes (animation) with camera key.
# Menu Item   :   Plugin->Camera Key Maker
# Usage       :   1 - Add keys  (Mayús.+ Q).
#             :   2 - Edit current keys (DOWN).
#             :   3 - Update current key edited (Mayús.+ Q).
#             :   4 - Preview animation (UP).
#             :   5 - Creating scenes   (ENTER).
#             :   6 - Previous/Next Key (Left/Right)
# Date        :   05/03/2011
# Type        :   Tool
#-----------------------------------------------------------------------------
=end

require 'sketchup.rb'
require 'bezier.rb'

$preview = false
$frames  = nil

class Camera_Key_Maker

 include Bezier
 include Math

 def initialize
  @points_target = nil
  @points_eye    = nil
  @old_eye       = nil
  @old_target    = nil
  @index         = nil
  @circle_target = nil
  @bottom        = nil
  @segment       = nil
  @selection     = nil
  @state         = nil
  @pt_to_move    = nil
  @pos_axis      = nil
  @pts_select    = nil
  @camLine       = nil
 end

 def activate

  @model    = Sketchup.active_model
  @entities = @model.active_entities
  @view     = @model.active_view
  @pages    = @model.pages
  @camera   = @view.camera
  @layers   = @model.layers
  @up       = Geom::Vector3d.new(0, 0, 1)
  @ip       = Sketchup::InputPoint.new

  # SET DEFAULT VALUES
  self.reset(@view)

  @my_layer_eye  = self.get_layer("eye_backup")
  @my_layer_tar  = self.get_layer("target_backup")

  # RETURN IF NOT SAME SIZE
  if (@points_eye.size > 0 && @points_target.size > 0)
   if (@points_eye.size != @points_target.size)
    UI.beep
    message  = "The paths should have the same number of points\r\n"
    message += "points eye = #{@points_eye.size} , points target = #{@points_target.size}"
    UI.messagebox message ,MB_OK
    @model.select_tool nil
   else
    @segment = @points_eye.size # * 2
    @bezier  = Bezier.points(@points_eye, @segment)
    Sketchup::set_status_text("Add Camera Key (Shift+Q)", SB_PROMPT)
   end
  else
   Sketchup::set_status_text("Place first Camera Key (Shift+Q)", SB_PROMPT)
  end

  @draw    = true
  @view.invalidate
  @model.start_operation("Camera Keys Maker")

 end # activate

 # RESET THE TOOL BACK TO IT INITIAL STATE
 def reset(view)
  # DEFAULT VALUES
  @state         = 0
  @index         = 0
  #@i             = 0
  @segment       = 10
  @drawn         = true
  $preview       = false
  @edit_cam      = false
  @points_target = []
  @points_eye    = []
  @bezier        = []
  @angle         = 0
  @pts_select     = nil
  # x axis is along the red line.
  # Y axis is along the green line.
  # Z axis is along the blue.
  # Corner == 0: top left, 1: top right, 2: bottom left,3: bottom right
  pointC  = view.corner 2
  @bottom = (pointC[1] - 40).to_i
  @camLine = [[10,@bottom + 12,0]\
  , [10,@bottom,0]\
  , [26,@bottom,0]\
  , [26,@bottom + 6,0]\
  , [34,@bottom,0]\
  , [34,@bottom + 12,0]\
  , [26,@bottom + 6,0]\
  , [26,@bottom + 12,0]\
  , [10,@bottom + 12,0] ]

  if view
   self.circle_target(view)
   view.tooltip = nil
   view.invalidate
  end

 end # reset

 # DESACTIVATE
 def deactivate(view)
  #view.invalidate if @drawn
  @ip            = nil
  @points_target = nil
  @points_eye    = nil
  @state         = nil
  @bezier        = nil
  @segment       = nil
  @bottom        = nil
  @index         = nil
  @up            = nil
  @pos_axis      = nil
  @pts_select    = nil
  view.invalidate  if view
  Sketchup.send_action("selectSelectionTool:")
 end # def

 # CIRCLE_TARGET
 def circle_target(view)
  # draw camera target circle
  pi_radian = PI/180
  ray = 20
  @circle_target = []
  c = view.center
  (0..24).each do |i|
   angle = 15 * i
   x = ray * (cos(angle * pi_radian)) + c.x
   y = ray * (sin(angle * pi_radian)) + c.y
   point = Geom::Point3d.new x, y, 0
   @circle_target.push point
  end
 end # circle_target

 # ONCANCEL
 def onCancel(flag, view)
  self.reset(view)
 end # onCancel

 # ONLBUTTONDOWN
 def onLButtonDown(flags, x, y, view)
  # Select the control point to move
  self.pick_point_to_move( x, y, view)
  @state = 1 if( @selection )
  @draw = true
  view.invalidate
 end # onLButtonDown


 # RETRIEVE VIEW AXIS
 def getAxis(view)
  pos_axis = nil
  a = view.camera.direction.to_a
  a.each_index { |i|
   if (a[i] == 1.0 or a[i] == -1.0)
    pos_axis = i;
    break;
   end }
   return pos_axis
  end

  # PICK_POINT_TO_MOVE
  def pick_point_to_move(x, y, view)

   old_pt_to_move = @pt_to_move
   @pos_axis      = getAxis(view)
   @selection     = nil

   if (@pos_axis )
    ph = view.pick_helper x, y
    path = nil

    if ph.pick_segment @points_eye
     path       = "eye"
     @selection = ph.pick_segment @points_eye
    elsif ph.pick_segment @points_target
     path       = "target"
     @selection = ph.pick_segment @points_target
    end

    if( @selection && path != nil )
     #puts "selection : #{@selection}"
     if path == "eye"
      @pt_to_move = @points_eye[@selection]
      @pts_select  = @points_eye
     else
      @pt_to_move = @points_target[@selection]
      @pts_select  = @points_target
     end
    else
     @pt_to_move = nil
    end
   end
   old_pt_to_move != @pt_to_move
  end # pick_point_to_move

  # CONSTRAINT AXIS
  def constraintAxis(view, pointAxis)
   point            = Geom::Point3d.new(0,0,0)
   point[@pos_axis] =  pointAxis[@pos_axis]
   plane  = [point, Geom::Vector3d.new(point.to_a)]
   vector = @ip.position.vector_to view.camera.eye.to_a
   Geom.intersect_line_plane([@ip.position, vector], plane)
  end

  # ONMOUSEMOVE
  def onMouseMove(flags, x, y, view)

   view.invalidate if not @drawn

   # Move the selected point if state = 1
   if(@state == 1 && @selection && @points_eye.size > 1)
    @ip.pick view, x, y
    return if not @ip.valid?

    if( @selection >= 0 )
     # Moving a control point
     point = @pt_to_move
     #the direction camera is aligned with the point to change
     if (@ip.position == view.camera.eye || point[@pos_axis] == 0.0)
      @pt_to_move = @ip.position
      @pt_to_move[@pos_axis] = point[@pos_axis]
     else
      @pt_to_move = constraintAxis(view, @pt_to_move)
     end
     @pts_select[@selection] = @pt_to_move
    end

    #@bezier = Bezier.points(@points_eye, @points_eye.size)
    @draw   = true
    view.invalidate

   else # state != 1 (no selection)
    if @points_eye.size == 0
     Sketchup::set_status_text("Place first Camera Key (Shift+Q), [Enter]=Set Scenes Creation, Arrays=[→]Next Key [←]Previous Key [↑]Animation Preview [↓]Edit Mode ", SB_PROMPT)
    else
     if  $preview == false
      if (@edit_cam == false)
       Sketchup::set_status_text("Add Camera Key #{@points_eye.size+1} (Shift+Q)", SB_PROMPT)
      else
       Sketchup::set_status_text("Update position of Camera Key #{@index+1} (Shift+Q)", SB_PROMPT)
      end # edit_cam
     end # preview
    end # eye.size == 0

   end # state = 1

  end # onMouseMove

  # ONLBUTTONUP
  def onLButtonUp(flags, x, y, view)
   view.invalidate
   return if not @state == 1
   @state = 0
   #backup path
   @selection = nil
   self.updatePoint()
   view.invalidate
  end # onLButtonUp

  # ------------------------------------
  # ONKEYDOWN (CAMERA MANUAL ANIMATION)
  # ------------------------------------
  def onKeyDown(key, x, y, view)

   # 63234 LEFT,  63235 RIGHT, 63232 UP, 63233 DOWN

    if  @points_eye.size > 0
     size = (@points_eye.size) -1

     #previous/next Key   <--[--1] || [++1]-->
      if key == 63234 or key == 63235

      if key == 63234 # LEFT <--
       if @index && @index > 0
        @index -= 1
        eye    = @points_eye.at(@index)
        target = @points_target.at(@index)
       else
        @index = size
        eye    = @points_eye.at(size)
        target = @points_target.at(size)
       end
      else # RIGHT -->
       if @index && @index < size
        @index += 1
        eye    = @points_eye.at(@index)
        target = @points_target.at(@index)
       else
        @index = 0
        eye    = @points_eye.at(0)
        target = @points_target.at(0)
       end
      end # key 65/63234

      @camera.set eye, target,  @up.clone
      Sketchup::set_status_text("Camera Key: #{@index+1}", SB_PROMPT)
      view.animation = nil
      @edit_cam    =  false
      @draw        =  true
      view.invalidate

     end # key 63234/63235

     # PLAY PREVIEW
     if key == 63232 # UP

      @edit_cam = false

      if @points_eye and @points_eye.size > 1
       $preview  = true
       view.animation = CameraAnimation.new(@points_eye, @points_target)
       #view.animation = cam_anim
       #place to last camera key
       @index    = (@points_eye.size) -1
      end
      #Animation stopped"
     end # 63232

     # EDIT current position camera key
     if key == 63233 # DOWN
      @edit_cam = true
      Sketchup::set_status_text("Edit Camera Key: #{@index+1} , Edit curve manually only Views [Top-Botton-Left-Right]", SB_PROMPT)
     end # key 63233

    end # @points_eye.size > 0

    @draw = true
    view.invalidate

   end #onKeyDown

   # ------------------------------------
   # ONKEYUP
   # ------------------------------------
   def onKeyUp(key, x, y, view)

    # ADD OR EDIT CAMERA : 81 == Mayús.+Q

    if key == 81 and $preview == false

     c = view.camera

     if c.is_a? Sketchup::Camera

      eye    = c.eye    if c.eye.is_a?    Geom::Point3d
      target = c.target if c.target.is_a? Geom::Point3d

      if @edit_cam == true
       @points_eye[@index]    = eye
       @points_target[@index] = target
       Sketchup::set_status_text("Camera Key #{@index} updated.", SB_PROMPT)
       @edit_cam = false
      else
       @points_eye.push    eye
       @points_target.push target
       Sketchup::set_status_text("Camera Key #{@points_eye.size} added.", SB_PROMPT)
      end

      #backup path
      self.updatePoint()

     end # end if Sketchup::Camera

     #backup paths && create scenes
    elsif key == 13 && $preview == false
     if (@points_eye.size >2)
      create_scene  = CreateScene.new(@points_eye, @points_target)
      create_scene.display_dialog()
     end
    end

    @draw = true
    view.invalidate

   end # onKeyUp

   #UPDATE POINT
   def updatePoint()
    if @points_eye.size >=2
     @segment = @points_eye.size * 2
     @bezier  = Bezier.points(@points_eye, @segment)
     #backup path
     self.make_bakup_path()
    end
   end

   # ONLBUTTONDOUBLECLICK: Place the target control point
   def onLButtonDoubleClick(key, x, y, view)
    # TODO : replace pick.point by direction of camera
    cam = view.camera
    @ip.pick view, x, y
    cam.set cam.eye, @ip.position , @up
    @draw = true
    view.invalidate
   end # onLButtonDoubleClick

   # DRAW
   def draw(view)

    view.dynamic = $preview
    legend = true #just to have the ability to disable

    if (legend == true)
     view.line_width    = 1
     view.drawing_color = "gray"
     view.line_stipple  = "."

     if $preview == false
      view.draw2d GL_LINE_LOOP, @circle_target
     end

     #display square of point edited
     if @state == 1
      if(@pt_to_move && getAxis(view))
       view.line_stipple = ""
       view.draw_points @pt_to_move, 15, 1, "blue"
      end

      #display line between eye and target
      if (@selection != nil)
       view.line_stipple = "."
       view.drawing_color = "gray"
       vector = ( @points_target[@selection] - @points_eye[@selection])
       vector.length  *= 100.0
       t = Geom::Transformation.translation vector
       pointTrans = @points_target[@selection].transform t
       view.draw_polyline(@points_eye[@selection], pointTrans)
      end

     end # state

     view.line_stipple = ""
     pointLineSmooth = [ [10,(@bottom - 10), 0] , [40,(@bottom - 10), 0]]
     pointLinePath1  = [ [10,(@bottom - 24), 0] , [40,(@bottom - 24), 0]]
     pointLinePath2  = [ [10,(@bottom - 29), 0] , [40,(@bottom - 29), 0]]

     view.drawing_color = "gray"
     view.draw2d GL_LINES, pointLinePath1
     view.draw2d GL_LINE_STRIP, @camLine

     view.drawing_color = "orange"
     view.draw2d GL_LINES, pointLinePath2

     view.line_stipple = "."
     view.drawing_color = "blue"
     view.draw2d GL_LINES, pointLineSmooth

     view.draw_text(Geom::Point3d.new( 50,(@bottom - 35), 0), "Path")
     view.draw_text(Geom::Point3d.new( 50,(@bottom - 20), 0), "Path smooth")
     view.draw_text(Geom::Point3d.new( 50,(@bottom), 0),      "Camera Key (#{@points_eye.size})")
    end

    if (@draw == true &&  $preview == false )

     view.line_width = 1
     if (@points_eye.size > 2)

      view.drawing_color = "blue"
      view.line_stipple = "."
      view.draw_polyline(@bezier)

      view.drawing_color = "gray"
      view.draw_line @points_eye[@index] ,  @points_target[@index]
      view.line_stipple = ""
      view.draw_polyline(@points_eye)
      view.drawing_color = "orange"
      view.draw_polyline(@points_target)

      #display the first point of path eye (start path)
      if (@points_eye[@index] != view.camera.eye)
       view.draw_points(@points_eye[@index], 15, 4, "red")
      end

      if (getAxis(view))
       view.draw_points(@points_eye,    4, 2, "gray")
       view.draw_points(@points_target, 4, 2, "orange")
      end

     end # bezier
    end # if draw

   end # draw

   # GETEXTENTS
   def getExtents
    bb = Geom::BoundingBox.new
    if (@points_eye.size > 0)
     bb.add @points_eye
     bb.add @points_target
    end
    bb
   end # getExtents

   # MAKE CAMERA PATH
   def make_bakup_path()

    #puts "DEF BACKUP PATH #{@points_eye.size}\n"
    if @points_eye.size >=2

     if @old_eye != nil
      if @old_eye.is_a? Sketchup::Curve
       @old_eye.edges.each{|e| @entities.erase_entities e}
      end

      if @old_target.is_a? Sketchup::Curve
       @old_target.edges.each{|e| @entities.erase_entities e}
      end
     end

     @old_eye    = nil
     @old_target = nil

     old_eye    = @entities.add_curve @points_eye
     old_target = @entities.add_curve @points_target
     #backup paths to back layer
     old_eye.each    { |e| e.layer = @my_layer_eye if e.is_a? Sketchup::Edge }
     old_target.each { |e| e.layer = @my_layer_tar if e.is_a? Sketchup::Edge }

     @old_eye    = old_eye[0].curve
     @old_target = old_target[0].curve
    end

   end # make_bakup_path

   # RETRIEVE EYE AND TARGET BACKUP IN LAYERs
   def get_layer(layer_name )

    layer_new = nil
    if @layers.length > 0

     # each layer (find layer name)
     @layers.each {| layer |

      if layer.name == layer_name
       layer_new = layer
       @model.active_layer = layer_new

       @entities.each { |entitie|
        if entitie.valid? and entitie.layer.name  == layer_name
         if entitie.typename == "Edge" and entitie.curve

          object = (entitie.curve.vertices.collect{|v| v.position})
          if layer_name    ==  "eye_backup"
           @old_eye        =   entitie.curve
           @points_eye     = object
          elsif layer_name ==  "target_backup"
           @points_target = object
           @old_target   =   entitie.curve
          end
          layer_new.visible = false

          #erase curves ?
          # @entities.erase_entities entitie.curve.edges
         end # entitie.typename
         break # loop
        end
       } # each

       @model.active_layer = @model.layers[0]

       break
      end
     }  # each layers
    end # layers length

    # add layer if not exist
    if not layer_new
     layer_new               =  @layers.add(layer_name)
     layer_new.page_behavior = LAYER_HIDDEN_BY_DEFAULT
     layer_new.visible       = false
    end

    return layer_new
   end # get_layer


  end # class Camera_Key_Maker

  class CreateScene

   def initialize(eye, target)
    @model        = Sketchup.active_model
    @entities    = @model.active_entities
    @view        = @model.active_view
    @pages       = @model.pages
    @shadowinfo  = @model.shadow_info
    @shade_time  = @shadowinfo['ShadowTime']

    @dialog     = nil
    @use_sun    = nil
    #@sun_start  = nil
    @offset     = nil
    @eye        = eye
    @target     = target
    @segment    = (@eye.size * 2 ) - 2

   end

   def display_dialog()

    @dialog  = UI::WebDialog.new( 'Set-Up Scenes', false, '', 440, 520, 200, 200, false )
    pathname = Sketchup.find_support_file "camera_path.html" ,"Plugins/CameraKeyMaker"

    @dialog.set_file( pathname )
    #show console call up
    @dialog.add_action_callback("showConsole") { |dialog, action|
    Sketchup.send_action("showRubyPanel:")}

    #default values
    @dialog.add_action_callback("onload") { |dialog, action|
     @dialog.execute_script("setNumPoints('#{@segment}')")
     @dialog.execute_script("setSunStart('#{@shade_time.strftime('%H.%M')}')")
     if $frames != nil
      @dialog.execute_script("setNumFrames('#{$frames}')")
     end }

     @dialog.add_action_callback( 'get_data' ) do |another_wd, data|
      if data == 'cancel'
       self.cancel( @dialog )
      elsif data == 'showConsole'
       Sketchup.send_action("showRubyPanel:")
      elsif data != nil
       self.make_board( data )
      end
     end

     #force dialog size
     @dialog.set_size 440,450
     @dialog.show_modal()

    end # def

    def make_board(data)

     #puts "\n-----DATA RECEIVED-----\n"
     values = eval(data)
     values.each do |key, value|

      # puts key + ' = ' + value
      case key
      when 'frame'
       $frames = value
      when 'use_sun'
       @use_sun = value
      #when 'sun_start'
       # @sun_start = value
      when 'offset'
       @offset = value
      end #case
      end #each

      #puts "frame : #{$frames}\nuse_sun : #{@use_sun}\noffset : #{@offset}\n"
      cancel( @dialog )
      if( $frames != nil)
       self.create_scene()
      end
     end

     def cancel( dialog )
      dialog.close()
     end

     def create_scene()

      #puts "CAM_Anim : creation de la camera"

      # Erase all pages if exist
      if @pages.size > 0
       pgs = Array.new
       @model.pages.each {|page| pgs.push page}
       pgs.each {|page| @model.pages.erase page}
      end

      up         = Geom::Vector3d.new(0, 0, 1)
      #time_start = @shade_time

      pts_eye    = Bezier.points(@eye,    @segment)
      pts_target = Bezier.points(@target, @segment)
      edge_length = 0.0
      offset_time = 0.0
      tmp_time    = 0.0
      frames_per_segment  = $frames.to_f

      if (@use_sun && @offset != nil)
       result      = @offset.split(/\./i)
       offset_time = ((result[0].to_i * (60*60)) + (result[1].to_i * 60)).to_f
      end

      generate_scene = true

      if generate_scene

       pts_eye.each_index do |index|

        newCamera      = Sketchup::Camera.new( pts_eye[index], pts_target[index], up)
        @view.camera   = newCamera
        newPage        = @pages.add
        shadowinfoPage = newPage.shadow_info
        newPage.delay_time = 0

        if (index < ((pts_eye.size) -1) )
         edge_length = (sprintf("%.3f", (pts_eye[index].distance  pts_eye[index+1]))).to_f
        end

        #transition_time
        if index == 0
         transition_time = 0
         tmp_time        =  @shade_time
        else
         tmp_time       += offset_time
         transition_time = (sprintf("%.3f", ((edge_length * frames_per_segment )/100))).to_f
        end

        shadowinfoPage['ShadowTime'] = tmp_time
        newPage.transition_time      = transition_time

       end # end each

      end # end generate true


      #go to the first page
      @pages.selected_page = @pages[0]
      #restore shade time
      @shadowinfo['ShadowTime'] = @shade_time

     end # create_scene

    end # class createScene

    class CameraAnimation

     # 1 microseconde = 0.1
     # 1 second       = 0.1 * 10   || 1.0   s
     # 10 second      = 0.1 * 100  || 10.0  s
     # 1 minute       = 0.1 * 600  || 60.0  s
     # 1 heure        = 60s * 60m  || 36000 s
     # time += 0.1    = add 1 microsecond

     include Bezier

     def initialize(eye, target)

      #puts "\n-------CameraAnimation-------\nInitialize\n"

      model        = Sketchup.active_model
      @shadowinfo  = model.shadow_info
      @shade_time  = @shadowinfo['ShadowTime']
      @back_shade  = @shadowinfo['ShadowTime']
      @up          = Geom::Vector3d.new(0, 0, 1)
      @startTime   =  Time.now
      #length_curve = 0
      @frame       = 0

      frames_per_second  = 15 # more frame = more smooth

      # (0..eye.size-2).each do |i|
       # length_curve += (eye[i].distance eye[i+1]).to_m
       # end

       @segment    = ((eye.size) -1) * (frames_per_second) # + (eye.size)-1 )
       @pts_eye    = Bezier.points(eye,    @segment)
       @pts_target = Bezier.points(target, @segment)

       @save_image = false
       date = Time.now.strftime("%d%m%Y")
       @fileName =  "c:/Temp/Anim/#{date}_img_#{@frame}.png"

      end #initialize

      def nextFrame(view)
       dateTime = Time.now
       totalTime = time_ms(@startTime, dateTime)
       #@shade_time += totalTime * 30

       fps = 1
       if( totalTime >= 0.01 )
        fps = (@frame / totalTime).to_i
       end

       eye    = @pts_eye.at(@frame)
       target = @pts_target.at(@frame)
       up = @up.clone
       view.camera.set(eye, target, up)

       if @save_image
        keys = {
         :filename => @fileName,
         :width => 320,
         :height => 240,
         :antialias => true,
         :compression => 0.9,
        :transparent => true }
        view.write_image keys
       end

       view.show_frame(0)
       @frame += 1
       elapsed = dateTime.to_f - @startTime.to_f
       mins, secs = elapsed.divmod 60.0
       total = ("%02d:%2d"%[mins.to_i, secs])
       Sketchup::set_status_text("fps : #{fps}   Segments : #{@segment}       Elapsed Time : #{total} (mm:ss)", SB_PROMPT)
       return @frame < @pts_eye.size

      end #nextFrame

      def stop
       #Restore ShadowTime
       @shadowinfo["ShadowTime"] =  @back_shade
       $preview = false
      end

      def time_ms(start, finish)
       # time.to_f = time in millisecond
       (sprintf("%.3f", (finish - start))).to_f
      end
     end # class
